#include "KiteVM.h"
#include "KiteIdentifier.h"
#include "KiteObject.h"

using namespace KiteVM;
using namespace std;

Instruction::Instruction(string f, int l)
: next(NULL), file(f), line(l)  { }

Instruction::~Instruction() { }

void Instruction::deallocate() {
	Instruction *ptr = next;
	while(ptr) {
		Instruction *cur = ptr;
		ptr = cur->next;
		delete cur;
	}
	delete this;
}

void Instruction::addToEnd(Instruction *p) {
	Instruction *cur = this;
    if (cur) {
        while(cur->next) cur = cur->next;
        cur->next = p;
    }
}

void Instruction::outTree() {
	Instruction *cur = this;
	while(cur) {
		std::cout << cur->toString() << std::endl;
		cur = cur->next;
	}
}

string processEscapes(string str) {
    string ret, replacement;
    int pos = 0, size;
    
    // Filter out leading/trailing "/'.
    ret = str.substr(1, str.size() - 2);
    
    // Process escapes
    while((pos = ret.find("\\", pos)) < ret.size()) {
        size = 2;
        switch(ret[pos+1]) {
        case 'n':
            replacement = "\n";
            break;
        case 'r':
            replacement = "\r";
            break;
        case '\\':
            replacement = "\\";
            break;
        case 't':
            replacement = "\t";
            break;
        case '"':
            replacement = "\"";
            break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9': {
            // ASCII escape
            char code[2];
            int i = 2;
            code[0] = ret[pos+1] - '0';
            while(isdigit(ret[pos+i])) {
                code[0] *= 10;
                code[0] += ret[pos+i] - '0';
                i++;
            }
            code[1] = '\0';
            replacement = code;
            size += i - 2;
            break;
        }
        case 'x': {
            // Hex escape
            char hcode[2];
            int j = 3;
            if (ret[pos+2] >= 'A' && ret[pos+2] <= 'F') {
                hcode[0] = (ret[pos+2] - 'A') + 10;
            } else if (ret[pos+2] >= 'a' && ret[pos+2] <= 'f') {
                hcode[0] = (ret[pos+2] - 'a') + 10;
            } else {
                hcode[0] = ret[pos+2] - '0';
            }
            while(isxdigit(ret[pos+j])) {
                hcode[0] *= 16;
                if (ret[pos+j] >= 'A' && ret[pos+j] <= 'F') {
                    hcode[0] += (ret[pos+j] - 'A') + 10;
                } else if (ret[pos+j] >= 'a' && ret[pos+j] <= 'f') {
                    hcode[0] += (ret[pos+j] - 'a') + 10;
                } else {
                    hcode[0] += ret[pos+j] - '0';
                }
                j++;
            }
            hcode[1] = '\0';
            replacement = hcode;
            size += j - 2;
            break;
        }
        };
        
        ret = ret.replace(pos, size, replacement);
        pos = pos + 1;
    }
    
    return ret;
}
