#include <sstream>
#include "KiteVM.h"
#include "KiteList.h"
#include "KiteString.h"
#include "KiteNumber.h"
#include "KiteReference.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"

using namespace KiteVM;
using namespace std;

KiteReference<KiteObject> KiteList::classinfo;

KiteReference<KiteObject> KiteList_opplus(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = (KiteList&)(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = (KiteList*)(param.getObj());
	} else {
		list2 = new KiteList(param);
	}
	
	KiteReference<KiteObject> ret(list1.merge(*list2));
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ret;
}

KiteReference<KiteObject> KiteList_intersect(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = (KiteList&)(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = (KiteList*)(param.getObj());
	} else {
		list2 = new KiteList(param);
	}

	KiteList *ret = new KiteList(false);
    
    // First, do the array items
    for (int i = 0; i < list1.getArray().size(); i++) {
        for (int j = 0; j < list2->getArray().size(); j++) {
            if (*list1.getArray().at(i) == *list2->getArray().at(j)) {
                ret->getArray().push_back(list1.getArray().at(i));
                break;
            }
        }
    }
    
    // Now do the dictionary items
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator k = list1.getItems().begin(), j;
	for ( ; k != list1.getItems().end(); k++) {
		if (list2->getItems().find((*k).first) != list2->getItems().end()) {
			if (valuesEqual(vm, (*k).second, (*list2->getItems().find((*k).first)).second)) {
				ret->getItems()[(*k).first] = (*k).second;
			}
		}
	}

	KiteReference<KiteObject> ref(ret);
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ref;
}

KiteReference<KiteObject> KiteList_except(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = (KiteList&)(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = (KiteList*)(param.getObj());
	} else {
		list2 = new KiteList(param);
	}

	KiteList *ret = new KiteList(false);

    // First, do the array items
    for (int i = 0; i < list1.getArray().size(); i++) {
        for (int j = 0; j < list2->getArray().size(); j++) {
            if (*list1.getArray().at(i) == *list2->getArray().at(j)) {
                goto bottom;
            }
        }
        ret->getArray().push_back(list1.getArray().at(i));
bottom: continue;
    }
    
    // Now do the dictionary items
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator k = list1.getItems().begin(), j;
	for ( ; k != list1.getItems().end(); k++) {
		if (list2->getItems().find((*k).first) == list2->getItems().end()) {
			if (valuesEqual(vm, (*k).second, (*list2->getItems().find((*k).first)).second)) {
				ret->getItems()[(*k).first] = (*k).second;
			}
		}
	}

	KiteReference<KiteObject> ref(ret);
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ref;
}

KiteReference<KiteObject> KiteList_opderef(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list = (KiteList&)(*thisObj);

	return list[param];
}

KiteList *KiteList::merge(KiteList &rhs) {
	KiteList *newlist = new KiteList(*this);

    // Insert all non-dictionary entries
    for (int i = 0; i < rhs.arrayItems.size(); i++) {
        newlist->arrayItems.push_back(rhs.arrayItems.at(i));
    }
    
	// Merge dictionary based items next.
	// Do not insert items with keys that already exist.
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = rhs.getItems().begin();
	for ( ; i != rhs.getItems().end(); i++) {
		newlist->dictItems[(*i).first] = (*i).second;
	}

	return newlist;
}

KiteReference<KiteObject> KiteList_string(VM &vm, KiteReference<KiteObject> param) {
	KiteList &lst = (KiteList&)(*THIS_OBJECT());
	stringstream str;

	str << "[";
    for (int i = 0; i < lst.getArray().size(); i++) {
        KiteReference<KiteObject> cur(lst.getArray().at(i));
        KiteReference<KiteObject> v(coerceString(vm, cur));
        KiteString &val = (KiteString&)(*v);
        str << val.getString() << " ";
    }
        
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator m = lst.getItems().begin();
	for ( ; m != lst.getItems().end(); m++) {
		KiteReference<KiteObject> k(coerceString(vm, (*m).first));
		KiteReference<KiteObject> v(coerceString(vm, (*m).second));
		KiteString &key = (KiteString&)(*k);
		KiteString &val = (KiteString&)(*v);
		str << key.getString() << ":" << val.getString() << " ";
	}
	str << "]";

	string ret = str.str();
	return KiteReference<KiteObject>(new KiteString(ret.erase(ret.size()-2, 1)));
}

KiteReference<KiteObject> KiteList_print(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> t(coerceString(vm, THIS_OBJECT()));
	KiteString &lst = (KiteString&)(*t);
	cout << lst.getString() << endl;
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteList_decide(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> to(THIS_OBJECT());
	KiteList &lst = (KiteList&)(*to);
	stringstream err;

	// Go through each condition. If true, execute the code object next door.
	int condnum = 1;
	for (int i = 0; i < lst.getArray().size(); i++) {
    	if (typeid(*lst.getArray().at(i)) != typeid(KiteBoolean)) {
			// Not a boolean expression
			err << "Condition " << condnum << " must evaluate to boolean value";
			return KiteReference<KiteObject>(new KiteTree(err.str()));
		}

		KiteBoolean &b = (KiteBoolean&)(*lst.getArray().at(i));
		if (b.boolVal()) {
			// Condition evaluates to true, execute code block.
			i++;
			KiteReference<KiteObject> code(lst.getArray().at(i));
			if (typeid(*code) == typeid(KiteCode)) {
				//vm.syms.removeContext();
				KiteReference<KiteObject> ret((*code).call(vm, param, false));
				//vm.syms.addContext();
				return ret;
			} else {
				return (lst.getArray().at(i));
			}
		}

		i++; // Skip past and go on to the next condition
	}

	// Return false; none of the conditions passed.
	return KiteReference<KiteObject>(new KiteBoolean(false));
}

KiteReference<KiteObject> KiteList_sum(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &lst = (KiteList&)(*thisObj);

	// Go through each element in the list
    KiteReference<KiteObject> ret(new KiteNumber(0));
    KiteReference<KiteObject> addFxn((*ret).dereference(vm, "op+"));
    for (int i = 0; i < lst.getArray().size(); i++) {
        vm.syms.addContext();
        vm.syms.addVar("this", ret);
        ret = (*addFxn).call(vm, coerceNumber(vm, lst.getArray().at(i)));
        vm.syms.removeContext();
    }
    
	// Return the sum.
    return ret;
}

KiteReference<KiteObject> KiteList_product(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &lst = (KiteList&)(*thisObj);

	// Go through each element in the list
    KiteReference<KiteObject> ret(new KiteNumber(1));
    KiteReference<KiteObject> addFxn((*ret).dereference(vm, "op*"));
    for (int i = 0; i < lst.getArray().size(); i++) {
        vm.syms.addContext();
        vm.syms.addVar("this", ret);
        ret = (*addFxn).call(vm, coerceNumber(vm, lst.getArray().at(i)));
        vm.syms.removeContext();
    }
    
	// Return the sum.
    return ret;
}

KiteReference<KiteObject> KiteList_length(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &lst = (KiteList&)(*thisObj);

    return KiteReference<KiteObject>(new KiteNumber((int)(lst.getItems().size() + lst.getArray().size())));
}

void KiteList::initialize() {
	syms["op+"] = KiteReference<KiteObject>(new KiteCode(KiteList_opplus));
	syms["opor"] = KiteReference<KiteObject>(new KiteCode(KiteList_opplus));
	syms["opand"] = KiteReference<KiteObject>(new KiteCode(KiteList_intersect));
	syms["op/"] = KiteReference<KiteObject>(new KiteCode(KiteList_except));
	syms["opderef"] = KiteReference<KiteObject>(new KiteCode(KiteList_opderef));
	syms["string"] = KiteReference<KiteObject>(new KiteCode(KiteList_string));
	syms["print"] = KiteReference<KiteObject>(new KiteCode(KiteList_print));
	syms["decide"] = KiteReference<KiteObject>(new KiteCode(KiteList_decide));
	syms["sum"] = KiteReference<KiteObject>(new KiteCode(KiteList_sum));
	syms["product"] = KiteReference<KiteObject>(new KiteCode(KiteList_product));
	syms["length"] = KiteReference<KiteObject>(new KiteCode(KiteList_length));
}
