#ifndef KITELIST_H
#define KITELIST_H

#include <map>
#include "KiteReference.h"
#include "KiteObject.h"
#include "KiteKeyValPair.h"
#include "KiteNumber.h"
#include "KiteIdentifier.h"

using namespace std;

struct list_sort {
	bool operator()(const KiteReference<KiteObject> &x1, const KiteReference<KiteObject> &x2) const {
		return *x1 < *x2;
	}
};

class KiteList : public KiteObject {
public:
	KiteList()
		: KiteObject(false) {
        inherits = KiteObject::getClassinfo();
        type = "list";
		initialize();
	}

    KiteList(bool initializeNow) : KiteObject(false), instance(true) {
        inherits = KiteList::getClassinfo();
        type = "list";
        if (initializeNow) initialize();
    }
    
	KiteList(const KiteList &obj) : KiteObject(false), instance(true), dictItems(obj.dictItems) {
        inherits = KiteList::getClassinfo();
        type = "list";
        
        vector<KiteReference<KiteObject> >::const_iterator i = obj.arrayItems.begin();
        for ( ; i != obj.arrayItems.end(); i++) {
            arrayItems.push_back(*i);
        }
		//initialize();
    }

	KiteList(KiteReference<KiteObject> obj) : KiteObject(false), instance(true) {
        inherits = KiteList::getClassinfo();
        type = "list";
		//initialize();
		if (typeid(*obj) == typeid(KiteKeyValPair)) {
			KiteKeyValPair a = (KiteKeyValPair&)(*obj);
			dictItems[a.first] = a.second;
		} else {
            arrayItems.push_back(obj);
        }
	}

	KiteList *merge(KiteList &rhs);

	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > &getItems() { return dictItems; }
    vector<KiteReference<KiteObject> > &getArray() { return arrayItems; }
    
	KiteReference<KiteObject> operator[](const KiteReference<KiteObject> &idx) {
        if (typeid(*idx) == typeid(KiteNumber)) {
            return arrayItems.at(((KiteNumber&)(*idx)).intValue());
        }
		return dictItems[idx];
	}
    
    virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteList();
        }
        return classinfo;
    }
private:
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > dictItems;
    vector<KiteReference<KiteObject> > arrayItems;
	bool instance;
	void initialize();
    
    static KiteReference<KiteObject> classinfo;
};

#endif // KITELIST_H
