#include <stdlib.h>
#include <sstream>
#include <time.h>
#include "KiteVM.h"
#include "KiteNumber.h"
#include "KiteString.h"
#include "KiteReference.h"
#include "KiteIdentifier.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteList.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteNumber::classinfo;

KiteReference<KiteObject> KiteNumber_opplus(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() + number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() + number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opminus(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() - number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() - number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opmultiply(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() * number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() * number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opdivide(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	if ((number2.isFloat() && number2.floatValue() == 0) ||
		(!number2.isFloat() && number2.intValue() == 0)) {
		// Divide by 0 attempted. Bad dog!
		return KiteReference<KiteObject>(new KiteTree("Division by zero"));
	}

	return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() / number2.floatValue()));
}

KiteReference<KiteObject> KiteNumber_opmod(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	if ((number2.isFloat() && number2.floatValue() == 0) ||
		(!number2.isFloat() && number2.intValue() == 0)) {
		// Divide by 0 attempted. Bad dog!
		return KiteReference<KiteObject>(new KiteTree("Division by zero"));
	}

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() % number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opleftshift(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() << number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_oprightshift(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() >> number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_print(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));

	if (number.isFloat()) {
		cout << number.floatValue() << endl;
	} else {
		cout << number.intValue() << endl;
	}

	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteNumber_oplessthan(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 < number2));
}

KiteReference<KiteObject> KiteNumber_oplessequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 < number2 || number1 == number2));
}

KiteReference<KiteObject> KiteNumber_opequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 == number2));
}

KiteReference<KiteObject> KiteNumber_opnotequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 == number2)));
}

KiteReference<KiteObject> KiteNumber_opgreaterequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 < number2)));
}

KiteReference<KiteObject> KiteNumber_opgreaterthan(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 < number2) && !(number1 == number2)));
}

KiteReference<KiteObject> KiteNumber_opand(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() & number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opor(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() | number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opxor(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = (KiteNumber&)(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() ^ number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opnot(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));

	return KiteReference<KiteObject>(new KiteNumber(!number1.intValue()));
}

KiteReference<KiteObject> KiteNumber_number(VM &vm, KiteReference<KiteObject> param) {
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteNumber_string(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = (KiteNumber&)(*coerceNumber(vm, THIS_OBJECT()));
	stringstream str;
	string retstr;

	if (number1.isFloat()) str << number1.floatValue();
	else str << number1.intValue();

	str >> retstr;
	return KiteReference<KiteObject>(new KiteString(retstr));
}

KiteReference<KiteObject> KiteNumber_range(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> from(coerceNumber(vm, THIS_OBJECT()));
    KiteList *ret = new KiteList(false);
    KiteReference<KiteNumber> cur;
    
    KiteReference<KiteObject> to(coerceNumber(vm, (typeid(*param) == typeid(KiteList) ? ((KiteList&)(*param)).getArray()[0] : param)));
    KiteReference<KiteObject> step;
    if (typeid(*param) == typeid(KiteList)) {
        step = ((KiteList&)(*param)).getArray()[1];
    } else {
        step = KiteReference<KiteObject>(new KiteNumber(1));
    }
    
    KiteReference<KiteObject> compareFxn((*from).dereference(vm, "op<="));
    KiteReference<KiteObject> addFxn((*from).dereference(vm, "op+"));
    while(1) {
        vm.syms.addContext();
        vm.syms.addVar("this", from);
        KiteReference<KiteObject> comp((*compareFxn).call(vm, to));
        if (!((KiteBoolean&)(*comp)).boolVal()) break;
        ret->getArray().push_back(from);
        from = (*addFxn).call(vm, step);
        vm.syms.removeContext();
    }
    
    return KiteReference<KiteObject>(ret);
}


KiteReference<KiteObject> KiteNumber_random(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> to(coerceNumber(vm, THIS_OBJECT()));
    static bool random_initialized = false;
    
    if (!random_initialized) {
        random_initialized = true;
        srand(time(NULL));
    }
    
    return KiteReference<KiteObject>(new KiteNumber((int)(rand() % ((KiteNumber&)(*to)).intValue())));
}

void KiteNumber::initialize() {
	syms["number"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_number));
	syms["op+"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opplus));
	syms["op-"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opminus));
	syms["op*"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opmultiply));
	syms["op/"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opdivide));
	syms["opmod"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opmod));
	syms["op<<"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opleftshift));
	syms["op>>"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oprightshift));

	syms["op<"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oplessthan));
	syms["op<="] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oplessequals));
	syms["op=="] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opequals));
	syms["op!="] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opnotequals));
	syms["op>="] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opgreaterequals));
	syms["op>"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opgreaterthan));

	syms["opand"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opand));
	syms["opor"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opor));
	syms["opnot"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opnot));
	syms["xor"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opxor));

	syms["string"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_string));
	syms["print"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_print));
    
    syms["range"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_range));
    syms["random"] = KiteReference<KiteObject>(new KiteCode(KiteNumber_random));
}
