#ifndef KITENUMBER_H
#define KITENUMBER_H

#include <string>
#include "KiteObject.h"

class KiteNumber : public KiteObject {
public:
	KiteNumber()
		: KiteObject(true) {
        type = "number";
		initialize();

		inherits = KiteObject::getClassinfo();
	}

	KiteNumber(std::string num) : KiteObject(false) {
        type = "number";
		inherits = KiteNumber::getClassinfo();

		this->instance = true;
		//initialize();

		if (num.find(".") != std::string::npos) {
			floatNumber = atof(num.c_str());
			floatingPt = true;
		} else {
			integer = atoi(num.c_str());
			floatingPt = false;
		}
	}

	KiteNumber(int num) : KiteObject(false) {
		inherits = KiteNumber::getClassinfo();
        type = "number";
		this->instance = true;
		//initialize();
		integer = num;
		floatingPt = false;
	}

	KiteNumber(double num) : KiteObject(true) {
		inherits = KiteNumber::getClassinfo();
        type = "number";
		this->instance = true;
		//initialize();
		floatNumber = num;
		floatingPt = true;
	}

	/*virtual KiteReference<KiteObject> call(VM &vm, KiteObject &param);
	virtual KiteReference<KiteObject> dereference(VM &vm, KiteIdentifier &id);*/

	virtual bool operator<(const KiteNumber &rhs) const {
		if (floatingPt && !rhs.floatingPt) {
			return floatNumber < (double)rhs.integer;
		} else if (!floatingPt && rhs.floatingPt) {
			return (double)integer < rhs.floatNumber;
		} else {
			return integer < rhs.integer;
		}
	}

	virtual bool operator==(const KiteNumber &rhs) const {
		if (floatingPt && !rhs.floatingPt) {
			return floatNumber  == (double)rhs.integer;
		} else if (!floatingPt && rhs.floatingPt) {
			return (double)integer == rhs.floatNumber;
		} else {
			return integer == rhs.integer;
		}
	}

	int intValue() { return (floatingPt ? (int)floatNumber : integer); }
	double floatValue() { return (floatingPt ? floatNumber : (double)integer); }
	bool isFloat() { return floatingPt; }

    virtual KiteReference<KiteObject> getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteNumber();
        }
        return classinfo;
    }
    
private:
	bool floatingPt;
	int integer;
	double floatNumber;
    static KiteReference<KiteObject> classinfo;
    
	void initialize();
};

#endif // KITENUMBER_H
