#include <sstream>
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteTree.h"
#include "KiteIdentifier.h"
#include "KiteList.h"

using namespace std;

bool KiteObject::initialized = false;
KiteReference<KiteObject> KiteObject::classinfo(NULL);

KiteReference<KiteObject> KiteObject::call(VM &vm, KiteReference<KiteObject> param, bool overwriteArgs) {
	// Only valid for code objects
	return KiteReference<KiteObject>(new KiteTree("Call attempted on uncallable object"));
}

KiteReference<KiteObject> KiteObject_opderef(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));	
}

KiteReference<KiteObject> KiteObject_opor(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opand(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opequals(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opnotequals(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opless(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_oplessequals(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opgreater(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opgreaterequals(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opleftshift(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_oprightshift(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opplus(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opminus(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opmultiply(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opdivide(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opmod(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_opnot(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteTree("Operation not valid on object"));
}

KiteReference<KiteObject> KiteObject_new(VM &vm, KiteReference<KiteObject> param) {
	// Default constructor
	return NULL;
}

KiteReference<KiteObject> KiteObject_number(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> strval(coerceString(vm, THIS_OBJECT()));
    
	return KiteReference<KiteObject>(coerceNumber(vm, strval));
}

KiteReference<KiteObject> KiteObject_string(VM &vm, KiteReference<KiteObject> param) {
    stringstream str;

    str << "[" << (*THIS_OBJECT()).type << "@" << (THIS_OBJECT()).getObj() << "]";
	return KiteReference<KiteObject>(new KiteString(str.str()));
}

KiteReference<KiteObject> KiteObject_print(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> strval(coerceString(vm, THIS_OBJECT()));
    
    cout << ((KiteString&)(*strval)).getString() << endl;
    return strval;
}

KiteReference<KiteObject> KiteObject_type(VM &vm, KiteReference<KiteObject> param) {
	return KiteReference<KiteObject>(new KiteString((*THIS_OBJECT()).type));
}

KiteObject::KiteObject() : instance(false), inherits(NULL), type("object") {
	initialize();
}

KiteObject::KiteObject(bool init) : instance(false), inherits(NULL), type("object") {
	if (init) initialize();
}

void KiteObject::initialize() {
    if (!instance && classinfo == NULL) {
	syms["opor"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opor));
	syms["opderef"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opderef));
	syms["opand"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opand));
	syms["op=="] = KiteReference<KiteObject>(new KiteCode(KiteObject_opequals));
	syms["op!="] = KiteReference<KiteObject>(new KiteCode(KiteObject_opnotequals));
	syms["op<"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opless));
	syms["op<="] = KiteReference<KiteObject>(new KiteCode(KiteObject_oplessequals));
	syms["op>"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opgreater));
	syms["op>="] = KiteReference<KiteObject>(new KiteCode(KiteObject_opgreaterequals));
	syms["op<<"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opleftshift));
	syms["op>>"] = KiteReference<KiteObject>(new KiteCode(KiteObject_oprightshift));
	syms["op+"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opplus));
	syms["op-"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opminus));
	syms["op*"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opmultiply));
	syms["op/"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opdivide));
	syms["opmod"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opmod));
	syms["op!"] = KiteReference<KiteObject>(new KiteCode(KiteObject_opnot));
	syms["number"] = KiteReference<KiteObject>(new KiteCode(KiteObject_number));
	syms["string"] = KiteReference<KiteObject>(new KiteCode(KiteObject_string));
    syms["type"] = KiteReference<KiteObject>(new KiteCode(KiteObject_type));
    syms["print"] = KiteReference<KiteObject>(new KiteCode(KiteObject_print));
	syms["new"] = KiteReference<KiteObject>(new KiteCode(KiteObject_new));
    //classinfo = this;
    }
}

KiteObject::KiteObject(VM &vm, KiteReference<KiteObject> param) : inherits(NULL), instance(true), type("object") {
	initialize();
}

KiteReference<KiteObject>
KiteObject::create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> ret(new KiteObject(vm, param));

	// Call constructor
	KiteReference<KiteObject> fun = (*ret).dereference(vm, "new");
	(void)(*fun).call(vm, param); // Call constructor

	return ret;
}

KiteObject::~KiteObject() {
	// Everything will be deconstructed by default
}

KiteReference<KiteObject> KiteObject::dereference(VM &vm, string id) {
	// Go through class hierachy, looking for object
	KiteObject *cur = this;

	while(cur) {
		if (cur->syms.count(id) > 0) {
			return cur->syms[id];
		}
		cur = &(*cur->inherits);
	}

	// We can't find the object, so time to create it.
	return syms[id];
}

bool compareIdentifiers(const KiteReference<KiteIdentifier> &r1, const KiteReference<KiteIdentifier> &r2) {
	/*cout << "comparing " << (*r1).getString() << " and " << (*r2).getString() 
		<< ": " << (*r1 < *r2) << endl;*/
	return *r1 < *r2;
}

